/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.connector.jms.activemq.embedded;

import java.io.PrintWriter;
import java.util.Set;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.JMSException;
import javax.resource.ResourceException;
import javax.resource.spi.ConnectionManager;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.ManagedConnection;
import javax.resource.spi.ManagedConnectionFactory;
import javax.resource.spi.ResourceAdapter;
import javax.resource.spi.ResourceAdapterAssociation;
import javax.security.auth.Subject;

import org.apache.commons.beanutils.PropertyUtils;
import org.bluestemsoftware.specification.eoa.ext.feature.jms.server.JMSServerFeature;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.server.Server;

public final class ManagedConnectionFactoryImpl implements ManagedConnectionFactory {

    private static final long serialVersionUID = 1L;

    private boolean isXA;
    private ConnectionManager connectionManager;
    private ManagedConnectionFactory underlyingManagedConnectionFactory;
    private PrintWriter logWriter;
    private Integer allPrefetchValues;
    private String clientid;
    private Integer durableTopicPrefetch;
    private Long initialRedeliveryDelay;
    private Integer inputStreamPrefetch;
    private Integer maximumRedeliveries;
    private String password;
    private Integer queueBrowserPrefetch;
    private Integer queuePrefetch;
    private Short redeliveryBackOffMultiplier;
    private Boolean redeliveryUseExponentialBackOff;
    private String serverUrl;
    private Integer topicPrefetch;
    private Boolean useInboundSession;
    private String userName;
    
    public ManagedConnectionFactoryImpl(boolean isXA) {
        this.isXA = isXA;
    }

    public Object createConnectionFactory() throws ResourceException {
        throw new ResourceException("not supported");
    }

    public Object createConnectionFactory(ConnectionManager connectionManager) throws ResourceException {
        this.connectionManager = connectionManager;
        return new LazyEmbeddedConnectionFactory();
    }

    public ManagedConnection createManagedConnection(Subject subject, ConnectionRequestInfo info) throws ResourceException {
        return underlyingManagedConnectionFactory.createManagedConnection(subject, info);
    }

    public PrintWriter getLogWriter() throws ResourceException {
        return underlyingManagedConnectionFactory.getLogWriter();
    }

    @SuppressWarnings("unchecked")
    public ManagedConnection matchManagedConnections(Set arg0, Subject arg1, ConnectionRequestInfo arg2) throws ResourceException {
        return underlyingManagedConnectionFactory.matchManagedConnections(arg0, arg1, arg2);
    }

    public void setLogWriter(PrintWriter logWriter) throws ResourceException {
        this.logWriter = logWriter;
    }

    public void setAllPrefetchValues(Integer allPrefetchValues) {
        this.allPrefetchValues = allPrefetchValues;
    }

    public void setClientid(String clientid) {
        this.clientid = clientid;
    }

    public void setDurableTopicPrefetch(Integer durableTopicPrefetch) {
        this.durableTopicPrefetch = durableTopicPrefetch;
    }

    public void setInitialRedeliveryDelay(Long initialRedeliveryDelay) {
        this.initialRedeliveryDelay = initialRedeliveryDelay;
    }

    public void setInputStreamPrefetch(Integer inputStreamPrefetch) {
        this.inputStreamPrefetch = inputStreamPrefetch;
    }

    public void setMaximumRedeliveries(Integer maximumRedeliveries) {
        this.maximumRedeliveries = maximumRedeliveries;
    }

    public void setPassword(String password) {
        this.password = password;
    }

    public void setQueueBrowserPrefetch(Integer queueBrowserPrefetch) {
        this.queueBrowserPrefetch = queueBrowserPrefetch;
    }

    public void setQueuePrefetch(Integer queuePrefetch) {
        this.queuePrefetch = queuePrefetch;
    }

    public void setRedeliveryBackOffMultiplier(Short redeliveryBackOffMultiplier) {
        this.redeliveryBackOffMultiplier = redeliveryBackOffMultiplier;
    }

    public void setRedeliveryUseExponentialBackOff(Boolean redeliveryUseExponentialBackOff) {
        this.redeliveryUseExponentialBackOff = redeliveryUseExponentialBackOff;
    }

    public void setServerUrl(String serverUrl) {
        this.serverUrl = serverUrl;
    }

    public void setTopicPrefetch(Integer topicPrefetch) {
        this.topicPrefetch = topicPrefetch;
    }

    public void setUseInboundSession(Boolean useInboundSession) {
        this.useInboundSession = useInboundSession;
    }

    public void setUserName(String userName) {
        this.userName = userName;
    }

    class LazyEmbeddedConnectionFactory implements ConnectionFactory {

        private ConnectionFactory underlyingConnectionFactory;

        public Connection createConnection() throws JMSException {
            if (underlyingConnectionFactory == null) {
                createUnderlying();
            }
            return underlyingConnectionFactory.createConnection();
        }

        public Connection createConnection(String userName, String password) throws JMSException {
            if (underlyingConnectionFactory == null) {
                createUnderlying();
            }
            return underlyingConnectionFactory.createConnection(userName, password);
        }

        private void createUnderlying() throws JMSException {

            Server server = SystemContext.getContext().getSystem().getServer();
            JMSServerFeature embeddedServer = server.getFeature(JMSServerFeature.class);

            // because activemq-ra resource adapter directly references core class
            // ActiveMQConnectionFactory, we will get a class cast exception if we
            // attempt to set class returned via getEmbeddedConnectionFactory
            // on ResourceAdapter, i.e. because they are loaded by diff classloaders.
            // so ... we must scope all activemq-ra classes to jms server deployment

            ConnectionFactory temp = embeddedServer.getEmbeddedConnectionFactory(isXA);

            // activemq-ra has no ActiveMQManagedXAConnectionFactory class as of
            // as of 4.0.2

            if (isXA) {
                throw new JMSException("Not yet supported.");
            }

            ClassLoader scopedLoader = temp.getClass().getClassLoader();

            try {
                Class<?> clazz = scopedLoader.loadClass("org.apache.activemq.ra.ActiveMQManagedConnectionFactory");
                underlyingManagedConnectionFactory = (ManagedConnectionFactory)clazz.newInstance();
                underlyingManagedConnectionFactory.setLogWriter(logWriter);
            } catch (Exception ex) {
                throw new JMSException("Failed creating connection to embedded jms server. " + ex);
            }

            ResourceAdapter resourceAdapter = null;
            try {
                Class<?> clazz = scopedLoader.loadClass("org.apache.activemq.ra.ActiveMQResourceAdapter");
                resourceAdapter = (ResourceAdapter)clazz.newInstance();
            } catch (Exception ex) {
                throw new JMSException("Failed creating connection to embedded jms server. " + ex);
            }

            try {
                PropertyUtils.setSimpleProperty(resourceAdapter, "allPrefetchValues", allPrefetchValues);
                PropertyUtils.setSimpleProperty(resourceAdapter, "clientid", clientid);
                
                PropertyUtils.setSimpleProperty(resourceAdapter, "userName", userName);
                PropertyUtils.setSimpleProperty(resourceAdapter, "password", password);
                PropertyUtils.setSimpleProperty(resourceAdapter, "serverUrl", serverUrl);
                PropertyUtils.setSimpleProperty(resourceAdapter, "useInboundSession", useInboundSession);
                //PropertyUtils.setSimpleProperty(resourceAdapter, "redeliveryBackOffMultiplier", redeliveryBackOffMultiplier);
                PropertyUtils.setSimpleProperty(resourceAdapter, "initialRedeliveryDelay", initialRedeliveryDelay);
                PropertyUtils.setSimpleProperty(resourceAdapter, "maximumRedeliveries", maximumRedeliveries);
                PropertyUtils.setSimpleProperty(resourceAdapter, "redeliveryUseExponentialBackOff", redeliveryUseExponentialBackOff);
                PropertyUtils.setSimpleProperty(resourceAdapter, "durableTopicPrefetch", durableTopicPrefetch);
                PropertyUtils.setSimpleProperty(resourceAdapter, "inputStreamPrefetch", inputStreamPrefetch);
                PropertyUtils.setSimpleProperty(resourceAdapter, "queueBrowserPrefetch", queueBrowserPrefetch);
                PropertyUtils.setSimpleProperty(resourceAdapter, "queuePrefetch", queuePrefetch);
                PropertyUtils.setSimpleProperty(resourceAdapter, "topicPrefetch", topicPrefetch);
            } catch (Exception ex) {
                throw new JMSException("Failed creating connection to embedded jms server."
                        + " Error configuring resourceAdapter. "
                        + ex.getMessage());
            }

            try {
                ((ResourceAdapterAssociation)underlyingManagedConnectionFactory)
                        .setResourceAdapter(resourceAdapter);
            } catch (ResourceException re) {
                throw new JMSException("Failed creating connection to embedded jms server."
                        + " Error configuring resourceAdapter. "
                        + re.getMessage());
            }

            try {
                underlyingConnectionFactory = (ConnectionFactory)underlyingManagedConnectionFactory
                        .createConnectionFactory(connectionManager);
            } catch (ResourceException re) {
                throw new JMSException("Failed creating connection to embedded jms server."
                        + " Error configuring ConnectionFactory. "
                        + re.getMessage());
            }

        }

    }

}
